// simple disassembler for microdata Micro 1600/21, Micro 821
// Armin Diehl <ad@ardiehl.de> 13 August 2015

/* 13.08.2015 ad: initial version
 * 08.10.2015 ad: allow comments as well 0x0a as line terminator in bf boot files 
 *  
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdint.h>
#include <ctype.h>
#include <getopt.h>

#define VERMAJ 1
#define VERMIN 0
#define MYSELF "Microdata 1600/21,821 disassembler %d.%02d, Armin Diehl <ad@ardiehl.de> October 8, 2015\n"

#define maxOpcdLen  4          // max opcode length (for building opcode table)
typedef char OpcdStr[maxOpcdLen+1];
struct OpcdRec
{
    OpcdStr         name;       // opcode name
    short           typ;        // opcode type
    unsigned short  opcodeval;
    short           varlen;     // parameter is based on word length
    char * cmnt;
};
typedef struct OpcdRec *OpcdPtr;

enum
{
    o_None,         // No operands
    o_BraRel,       // short branch
    o_Bra16,        // long branch
    o_Imm8,         // one byte as parameter
    o_Imm16,        // two bytes as parameter
    o_m0,           // m=0: 8 bit addr in page 0
    o_m1,           // m=1: relative 8 bit
    o_m2,           // m=2: indirect page 0
    o_m3,           // m=3: indirect relative
    o_m4,           // m=4: indexed
    o_m5,           // m=5: indexed with bias
    o_m6,           // m=6: extended address
    o_m7,           // m=7: indirect extended address or variable length imm
    o_jm0,           // m=0: 8 bit addr in page 0
    o_jm1,           // m=1: relative 8 bit
    o_jm2,           // m=2: indirect page 0
    o_jm3,           // m=3: indirect relative
    o_jm4,           // m=4: indexed
    o_jm5,           // m=5: indexed with bias
    o_jm6,           // m=6: extended address
    o_jm7,           // m=7: indirect extended address or variable length imm
    o_inout,
    o_IBM_OBM};

#define OP_JMP 0x60
#define OP_RETJMP 0x68

struct OpcdRec MD1600_opcdTab[] =
{
    {"HLT", o_None,     0x00, 0, "Halt cpu"},
    {"TRP", o_None,     0x01, 0, "Trap 16 Bit JMP [80], saves P to [80,81]"},
    {"?02", o_None,     0x02, 0},
    {"ESW", o_None,     0x03, 0, "Load sense switches to A(12-15)"},
    {"DIN", o_None,     0x04, 0, "Disable interrupts"},
    {"EIN", o_None,     0x05, 0, "Enable interrupts"},
    {"DRT", o_None,     0x06, 0, "Disable realtime clock"},
    {"ERT", o_None,     0x07, 0, "Enable realtime clock"},
    {"RO1", o_None,     0x08, 0, "Reset overflow and word length 1"},
    {"RO2", o_None,     0x09, 0, "Reset overflow and word length 2"},
    {"RO3", o_None,     0x0A, 0, "Reset overflow and word length 3"},
    {"RO4", o_None,     0x0B, 0, "Reset overflow and word length 4"},
    {"SO1", o_None,     0x0C, 0, "Set overflow and word length 1"},
    {"SO2", o_None,     0x0D, 0, "Set overflow and word length 2"},
    {"SO3", o_None,     0x0E, 0, "Set overflow and word length 3"},
    {"SO4", o_None,     0x0F, 0, "Set overflow and word length 4"},
    
    {"JOV", o_BraRel,   0x10, 0, "Jump if overlow set"},
    {"JAZ", o_BraRel,   0x11, 0, "Jump if A = 0"},
    {"JBZ", o_BraRel,   0x12, 0, "Jump if B = 0"},
    {"JXZ", o_BraRel,   0x13, 0, "Jump if X = 0"},
    {"JAN", o_BraRel,   0x14, 0, "Jump if A < 0"},
    {"JXN", o_BraRel,   0x15, 0, "Jump if X < 0"},
    {"JAB", o_BraRel,   0x16, 0, "Jump if A = B"},
    {"JAX", o_BraRel,   0x17, 0, "Jump if A = X"},
    {"NOV", o_BraRel,   0x18, 0, "Jump if no overflow"},
    {"NAZ", o_BraRel,   0x19, 0, "Jump if A <> 0"},
    {"NBZ", o_BraRel,   0x1A, 0, "Jump if B <> 0"},
    {"NXZ", o_BraRel,   0x1B, 0, "Jump if X <> 0"},
    {"NAN", o_BraRel,   0x1C, 0, "Jump if A >= 0"},
    {"NXN", o_BraRel,   0x1D, 0, "Jump if X >= 0"},
    {"NAB", o_BraRel,   0x1E, 0, "Jump if A <> B"},
    {"NAX", o_BraRel,   0x1F, 0, "Jump if A <> X"},
    
    /* shifts */
    {"LLA", o_Imm8,     0x20, 0, "Shift A left, bits shifted out of A15 are shifted into A0"},
    {"LLB", o_Imm8,     0x21, 0, "Shift B left, bits shifted out of A15 are shifted into A0"},
    {"LLL", o_Imm8,     0x22, 0, "Shift AB left, bits shifted out of B15 are shifted into A0"},
    {"?23", o_None,     0x23, 0, "UNKNOWN"},
    {"LRA", o_Imm8,     0x24, 0, "Shift A right, bits shifted out are lost"},
    {"LRB", o_Imm8,     0x25, 0, "Shift B right, bits shifted out are lost"},
    {"LRL", o_Imm8,     0x26, 0, "Shift AB right, bits shifted out are lost"},
    {"?27", o_None,     0x27, 0},
    {"ALA", o_Imm8,     0x28, 0, "Shift A left, bits shifted out are lost"},
    {"ALB", o_Imm8,     0x29, 0, "Shift B left, bits shifted out are lost"},
    {"ALL", o_Imm8,     0x2A, 0, "Shift AB left, bits shifted out are lost"},
    {"?2B", o_Imm8,     0x2B, 0},
    {"ARA", o_Imm8,     0x2C, 0, "Shift A right, A15 remains unchanged"},
    {"ARB", o_Imm8,     0x2D, 0, "Shift B right, A15 remains unchanged"},
    {"ARL", o_Imm8,     0x2E, 0, "Shift AB right, A15 remains unchanged"},
    {"?2F", o_None,     0x2F, 0},
    {"?30", o_None,     0x30, 0},
    
    
    /* input / output operations */
    {"IBA", o_inout,     0x31, 0, "Input byte to A"},
    {"IBB", o_inout,     0x32, 0, "Input byte to B"},
    {"IBM", o_IBM_OBM,  0x33, 0, "Input byte to memory"},
    {"NOP", o_None,     0x34, 0},
    {"CLC", o_None,     0x35, 0, "compare string (A) with (X) up to addr (of X) in B"},
    {"?36", o_None,     0x36, 0},
    {"?37", o_None,     0x37, 0},
    {"?38", o_None,     0x38, 0},
    
    {"OBA", o_inout,     0x39, 0, "output byte from A"},
    {"OBB", o_inout,     0x3A, 0, "output byte from B"},
    {"OBM", o_IBM_OBM,  0x3B, 0, "output byte from memory"},
    {"DAD", o_None,     0x3C, 0, "Decimal add, (X)=(X)+(B), variable, len in A"},
    {"DSB", o_None,     0x3D, 0, "Decimal sub, (X)=(X)-(B), variable, len in A"},
    
    {"MUL", o_Imm16,    0x3E, 0, "AB = A * param[16]"},
    {"DIV", o_Imm16,    0x3F, 0, "B=AB/param[16],A=AB MOD param[16]"},
    
    /* register operate */
    {"ORA", o_None,     0x40, 0, "A=A or B"},
    {"XRA", o_None,     0x41, 0, "A=A xor B"},
    {"ORB", o_None,     0x42, 0, "B=A or B"},
    {"XRB", o_None,     0x43, 0, "B=A xor B"},
    {"INX", o_None,     0x44, 0, "increment X"},
    {"DCX", o_None,     0x45, 0, "decrement X"},
    {"AWX", o_None,     0x46, 0, "add current word length to X"},
    {"SWX", o_None,     0x47, 0, "substract current word length from X"},
    {"INA", o_None,     0x48, 0, "Increment A"},
    {"INB", o_None,     0x49, 0, "Increment B"},
    {"OCA", o_None,     0x4A, 0, "Ones complement A, A = not A"},
    {"ORB", o_None,     0x4B, 0, "Ones complement B, B = not B"},
    {"TAX", o_None,     0x4C, 0, "Transfer A to X, X = A"},
    {"TBX", o_None,     0x4D, 0, "Transfer B to X, X = B"},
    {"TXA", o_None,     0x4E, 0, "Transfer X to A, A = X"},
    {"TXB", o_None,     0x4F, 0, "Transfer X to B, B = X"},
    
    /* Stack */
    {"RTN", o_None,     0x50, 0, "Return: pop O/W[8],P,B,A,X (9 bytes) and jump to P"},
    {"CAL", o_Bra16,    0x51, 0, "Call: push X,A,B,P,O/W[8] (9 bytes) and jump"},
    {"PLX", o_None,     0x52, 0, "Pull X from stack"},
    {"PSX", o_None,     0x53, 0, "Push X to stack"},
    {"PLA", o_None,     0x54, 0, "Pull A from stack"},
    {"PSA", o_None,     0x55, 0, "Push A to stack"},
    {"PLB", o_None,     0x56, 0, "Pull B from stack"},
    {"PSB", o_None,     0x57, 0, "Push B to stack"},
    {"MST", o_Imm16,    0x58, 0, "Mult Step"},
    {"ADX", o_Imm16,    0x59, 0, "Add to X"},
    {"JEP", o_BraRel,   0x5A, 0, "Jump if even parity (A contains even number of 1 bits"},
    {"EBX", o_None,     0x5B, 0, "Exchange B and X"},
    
    /* character/string manipulation */
    
    {"MOV",   o_None,     0x5C, 0, "move (X) to (A), end address (of X) in B"},
    {"GCC",   o_None,     0x5D, 0, "generate 16 bit CRC in A, (X)=src,B=end addr"},
    {"SCH",   o_None,     0x5E, 0, "Search"},
    {"GAP",   o_None,     0x5F, 0, "Gernerate ascii parity, src=(X), end addr=B"},
    
    /* Memory reference ops */
    {"JMP",   o_jm0,  0x60, 0, "Jump"},
    {"JMP",   o_jm1,  0x61, 0},
    {"JMP",   o_jm2,  0x62, 0},
    {"JMP",   o_jm3,  0x63, 0},
    {"JMP",   o_jm4,  0x64, 0},
    {"JMP",   o_jm5,  0x65, 0},
    {"JMP",   o_jm6,  0x66, 0},
    {"JMP",   o_jm7,  0x67, 0},
    
    {"RTJ",  o_jm0,  0x68, 0, "Return Jump, save retAaddr at effective addr and jump to the eff. addr+2"},
    {"RTJ",  o_jm1,  0x69, 0},
    {"RTJ",  o_jm2,  0x6A, 0},
    {"RTJ",  o_jm3,  0x6B, 0},
    {"RTJ",  o_jm4,  0x6C, 0},
    {"RTJ",  o_jm5,  0x6D, 0},
    {"RTJ",  o_jm6,  0x6E, 0},
    {"RTJ",  o_jm7,  0x6F, 0},
    
    {"IWM",  o_m0,  0x70, 0, "Increment word in memory"},
    {"IWM",  o_m1,  0x71, 0},
    {"IWM",  o_m2,  0x72, 0},
    {"IWM",  o_m3,  0x73, 0},
    {"IWM",  o_m4,  0x74, 0},
    {"IWM",  o_m5,  0x75, 0},
    {"IWM",  o_m6,  0x76, 0},
    {"IWM",  o_m7,  0x77, 0},
    
    
    {"DWM",  o_m0,  0x78, 0, "Decrement word in memory"},
    {"DWM",  o_m1,  0x79, 0},
    {"DWM",  o_m2,  0x7A, 0},
    {"DWM",  o_m3,  0x7B, 0},
    {"DWM",  o_m4,  0x7C, 0},
    {"DWM",  o_m5,  0x7D, 0},
    {"DWM",  o_m6,  0x7E, 0},
    {"DWM",  o_m7,  0x7F, 0},
    
    {"LDX",  o_m0,  0x80, 0, "Load 16 bit word to X"},
    {"LDX",  o_m1,  0x81, 0},
    {"LDX",  o_m2,  0x82, 0},
    {"LDX",  o_m3,  0x83, 0},
    {"LDX",  o_m4,  0x84, 0},
    {"LDX",  o_m5,  0x85, 0},
    {"LDX",  o_m6,  0x86, 0},
    {"LDX",  o_m7,  0x87, 0},
    
    {"STX",  o_m0,  0x88, 0, "Store X to memory"},
    {"STX",  o_m1,  0x89, 0},
    {"STX",  o_m2,  0x8A, 0},
    {"STX",  o_m3,  0x8B, 0},
    {"STX",  o_m4,  0x8C, 0},
    {"STX",  o_m5,  0x8D, 0},
    {"STX",  o_m6,  0x8E, 0},
    {"STX",  o_m7,  0x8F, 0},
    
    {"LDB",  o_m0,  0x90, 0, "Load 16 bit word to X"},
    {"LDB",  o_m1,  0x91, 0},
    {"LDB",  o_m2,  0x92, 0},
    {"LDB",  o_m3,  0x93, 0},
    {"LDB",  o_m4,  0x94, 0},
    {"LDB",  o_m5,  0x95, 0},
    {"LDB",  o_m6,  0x96, 0},
    {"LDB",  o_m7,  0x97, 0},
    
    {"STB",  o_m0,  0x98, 0, "Store B to memory"},
    {"STB",  o_m1,  0x99, 0},
    {"STB",  o_m2,  0x9A, 0},
    {"STB",  o_m3,  0x9B, 0},
    {"STB",  o_m4,  0x9C, 0},
    {"STB",  o_m5,  0x9D, 0},
    {"STB",  o_m6,  0x9E, 0},
    {"STB",  o_m7,  0x9F, 0},
    
    {"ADA",  o_m0,  0xA0, 0, "Add to A"},
    {"ADA",  o_m1,  0xA1, 0},
    {"ADA",  o_m2,  0xA2, 0},
    {"ADA",  o_m3,  0xA3, 0},
    {"ADA",  o_m4,  0xA4, 0},
    {"ADA",  o_m5,  0xA5, 0},
    {"ADA",  o_m6,  0xA6, 0},
    {"ADA",  o_m7,  0xA7, 0},
    
    {"ADV",  o_m0,  0xA8, 1, "Add variable to A or AB (word length)"},
    {"ADV",  o_m1,  0xA9, 1},
    {"ADV",  o_m2,  0xAA, 1},
    {"ADV",  o_m3,  0xAB, 1},
    {"ADV",  o_m4,  0xAC, 1},
    {"ADV",  o_m5,  0xAD, 1},
    {"ADV",  o_m6,  0xAE, 1},
    {"ADV",  o_m7,  0xAF, 1},
    
    {"SBA",  o_m0,  0xB0, 0, "Sub from A"},
    {"SBA",  o_m1,  0xB1, 0},
    {"SBA",  o_m2,  0xB2, 0},
    {"SBA",  o_m3,  0xB3, 0},
    {"SBA",  o_m4,  0xB4, 0},
    {"SBA",  o_m5,  0xB5, 0},
    {"SBA",  o_m6,  0xB6, 0},
    {"SBA",  o_m7,  0xB7, 0},
    
    {"SBV",  o_m0,  0xB8, 1, "Sub variable from A or AB (word length)"},
    {"SBV",  o_m1,  0xB9, 1},
    {"SBV",  o_m2,  0xBA, 1},
    {"SBV",  o_m3,  0xBB, 1},
    {"SBV",  o_m4,  0xBC, 1},
    {"SBV",  o_m5,  0xBD, 1},
    {"SBV",  o_m6,  0xBE, 1},
    {"SBV",  o_m7,  0xBF, 1},
    
    {"CPA",  o_m0,  0xC0, 0, "Compare A with 16 bit word"},
    {"CPA",  o_m1,  0xC1, 0},
    {"CPA",  o_m2,  0xC2, 0},
    {"CPA",  o_m3,  0xC3, 0},
    {"CPA",  o_m4,  0xC4, 0},
    {"CPA",  o_m5,  0xC5, 0},
    {"CPA",  o_m6,  0xC6, 0},
    {"CPA",  o_m7,  0xC7, 0},

    {"CPV",  o_m0,  0xC8, 1, "Compare A or AB with given data (variable length)"},
    {"CPV",  o_m1,  0xC9, 1},
    {"CPV",  o_m2,  0xCA, 1},
    {"CPV",  o_m3,  0xCB, 1},
    {"CPV",  o_m4,  0xCC, 1},
    {"CPV",  o_m5,  0xCD, 1},
    {"CPV",  o_m6,  0xCE, 1},
    {"CPV",  o_m7,  0xCF, 1},

    {"ANA",  o_m0,  0xD0, 0, "16 Bit AND"},
    {"ANA",  o_m1,  0xD1, 0},
    {"ANA",  o_m2,  0xD2, 0},
    {"ANA",  o_m3,  0xD3, 0},
    {"ANA",  o_m4,  0xD4, 0},
    {"ANA",  o_m5,  0xD5, 0},
    {"ANA",  o_m6,  0xD6, 0},
    {"ANA",  o_m7,  0xD7, 0},
    
    {"ANV",  o_m0,  0xD8, 1, "A or AB, AND variable length"},
    {"ANV",  o_m1,  0xD9, 1},
    {"ANV",  o_m2,  0xDA, 1},
    {"ANV",  o_m3,  0xDB, 1},
    {"ANV",  o_m4,  0xDC, 1},
    {"ANV",  o_m5,  0xDD, 1},
    {"ANV",  o_m6,  0xDE, 1},
    {"ANV",  o_m7,  0xDF, 1},
    
    {"LDA",  o_m0,  0xE0, 0, "Load 16 bit into A"},
    {"LDA",  o_m1,  0xE1, 0},
    {"LDA",  o_m2,  0xE2, 0},
    {"LDA",  o_m3,  0xE3, 0},
    {"LDA",  o_m4,  0xE4, 0},
    {"LDA",  o_m5,  0xE5, 0},
    {"LDA",  o_m6,  0xE6, 0},
    {"LDA",  o_m7,  0xE7, 0},
    
    {"LDV",  o_m0,  0xE8, 1, "load variable length into A or AB"},
    {"LDV",  o_m1,  0xE9, 1},
    {"LDV",  o_m2,  0xEA, 1},
    {"LDV",  o_m3,  0xEB, 1},
    {"LDV",  o_m4,  0xEC, 1},
    {"LDV",  o_m5,  0xED, 1},
    {"LDV",  o_m6,  0xEE, 1},
    {"LDV",  o_m7,  0xEF, 1},
    
    {"STA",  o_m0,  0xF0, 0, "Save A (16 bit) to given address"},
    {"STA",  o_m1,  0xF1, 0},
    {"STA",  o_m2,  0xF2, 0},
    {"STA",  o_m3,  0xF3, 0},
    {"STA",  o_m4,  0xF4, 0},
    {"STA",  o_m5,  0xF5, 0},
    {"STA",  o_m6,  0xF6, 0},
    {"STA",  o_m7,  0xF7, 0},
    
    {"STV",  o_m0,  0xF8, 0, "Store variable A or AB to given address"},
    {"STV",  o_m1,  0xF9, 0},
    {"STV",  o_m2,  0xFA, 0},
    {"STV",  o_m3,  0xFB, 0},
    {"STV",  o_m4,  0xFC, 0},
    {"STV",  o_m5,  0xFD, 0},
    {"STV",  o_m6,  0xFE, 0},
    {"STV",  o_m7,  0xFF, 0}};

// -------------------------------------------------------------------

#define fmtBF  0
#define fmtBin 1

int inputFmt = fmtBF;
int org = 0;
uint16_t currAddr = 0;
FILE *inFile;
int atEof;
int errors = 0;
unsigned char currByte;
int nextIsOrg;
char prefix[255],command[255],params[255],comment[255];
int currWordLength = 1;
int comments = 1;
int linesOut = 0;
int oldFmt = 0;
int assemblerOut = 0;


void strupper (char * s) {
    while (*s) {
        *s = toupper(*s);
        s++;
    }
}

void usage (char *prgName) {
    printf( \
      "usage: %s [-b] [-o OrgAddr] [-1] [-2] [-3] [-4] FileName\n" \
      " -b\n" \
      " --binary        binary file as input, default is mai basic four console boot file\n" \
      " -o addr\n" \
      " --address=addr  start address (hex) for binary input\n" \
      " -n\n" \
      " --nocomments    omit comments in output\n" \
      " -1              set word length to 1 (for ADV,SBV,CPV,LDV.. with m=7)\n" \
      " -2, -3, -4      set word length to 2, 3 or 4\n" \
      " -l\n"\
      " --old           old opcode format\n" \
      " -a\n" \
      " --asmout        omit address and hexdump of command\n" \
      "\n"
      "Basic four console boot file is an ascii file with upper case hex chars. Lines are\n"
      "terminated with 0x1f. First four bytes of a line is the 16 bit target address. A line\n"
      "with a target address only starts execution at target address.\n" \
      "Disassembly for variable opcodes (e.g. LDV) will fail if the code used diffetent\n"
      "word lengths since the opcode is the same and the word length can be set in code.\n" \
    
      ,prgName);
    _exit (1);
}

int hex2int (char * c) {
    int res;
    
    res = (int)strtol(c, NULL, 16);
    return res;
}

unsigned char nextChar() {
    unsigned char res = 0;
    int bytesRead;
    if (!atEof) {
        bytesRead = fread(&res,1,1,inFile);
        if (bytesRead != 1) { atEof = 1; res = 0; }
    }
    if ((res == 0x1f) || (res == 0x0a)) nextIsOrg = 1;
    return res;
}

unsigned char nextByte() {
    char hexStr[3];
    char res;
    if (inputFmt == fmtBin) {
        return nextChar();
    } else {
        if (nextIsOrg) return 0;
        hexStr[0] = nextChar(); 
        if (nextIsOrg) return 0;
        hexStr[1] = nextChar(); hexStr[2] = 0;
        if (nextIsOrg) return 0;
        res = hex2int(hexStr);
       // printf(" .. %02x ",res);
        return res;
    }
}


int8_t geti8() {
    char s[10]; int i;
    
    currAddr++;
    i = (int8_t)nextByte();
    sprintf(s,"%02x",(uint8_t)i); strcat(prefix,s);
    return i;
}

uint8_t get8() {
    int i;
    char s[10];
    
    currAddr++;
    i = (uint8_t)nextByte();
    sprintf(s,"%02x",i); strcat(prefix,s);
    return i;
}

uint16_t get16() {
    int i;
    i = get8() << 8;
    i = i + get8();
    return i;
}
    

void processOpcode() {
    int typ;
    int wordLength;
    int param,param2,param1;
    char cmd;
    int8_t relOfs;
    char stmp[20];
    char addrInfo[20];
    
    addrInfo[0] = 0;
    typ = MD1600_opcdTab[(short)currByte].typ;
    sprintf(prefix,"%04x %02x",currAddr,currByte);
    currAddr++;
    strcpy(command,MD1600_opcdTab[(short)currByte].name);
    params[0] = 0;
    comment[0] = 0;
    if (comments) {
      param = currByte;
      if (param >= 0x60) param &= 0xf8;
      if (MD1600_opcdTab[param].cmnt) {
        sprintf(comment,"; %s",MD1600_opcdTab[param].cmnt);
        if (currByte >= 0x60) {
            sprintf(stmp," (m=%d)",currByte & 0x07);
            strcat(comment,stmp);
        }
      }
    }

    
    
    cmd = currByte;
    
    switch (typ) {
        
        case o_None:         // No operands
          break;
        case o_BraRel:       // short branch
          relOfs = geti8();
          sprintf(params,"%04xH",currAddr+relOfs);
          sprintf(addrInfo," (ofs: %d=0x%02x)",relOfs,relOfs);
          break;
        case o_Bra16:        // long branch
          param = get16();
          sprintf(params,"%04xH",param);
          break;
        case o_Imm8:         // one byte as parameter
          param = get8();
          if (param < 0x0a)
            sprintf(params,"%d",param);
          else 
            sprintf(params,"%02xH",param);
          break;
        case o_Imm16:        // two bytes as parameter
          param = get16();
          sprintf(params,"%04xH",param);
          break;
        case o_m0:           // m=0: 8 bit addr in page 0
          param = get8();
          if (oldFmt)
            sprintf(params,"%02xH",param);
          else
            sprintf(params,"[%02xH]",param);
          break;
        case o_jm0:           // m=0: 8 bit addr in page 0
          param = get8();
          if (oldFmt)
            sprintf(params,"X'%02x'",param);
          else
            sprintf(params,"SHORT %02xH",param);
          break;
        case o_m1:           // m=1: relative 8 bit
          relOfs = geti8();
          if (oldFmt)
            sprintf(params,"X'%04x'",currAddr+relOfs);
          else
            sprintf(params,"SHORT [%04xh]",currAddr+relOfs/*-2*/);
            sprintf(addrInfo," (ofs: %d=0x%02x)",relOfs,relOfs);
          break;
        case o_jm1:           // m=1: relative 8 bit
          relOfs = geti8();
          if (oldFmt) {
            sprintf(params,"X'%04x'",currAddr+relOfs/*-2*/);
          } else
            sprintf(params,"SHORT %04xH",currAddr+relOfs/*-2*/);
          sprintf(addrInfo," (ofs: %d=0x%02x)",relOfs,relOfs);
          break;
        case o_m2:           // m=2: indirect page 0
          //strcat(command,"'");
          param = get8();
          if (oldFmt) {
			strcat(command,"'");
            sprintf(params,"%02x",param);
          } else
            sprintf(params,"[*%02xH]",param);
          break;
        case o_jm2:           // m=2: indirect page 0
          param = get8();
          sprintf(params,"[%02xH]",param);
          break;
        case o_m3:           // m=3: indirect relative
          relOfs = geti8();
          if (oldFmt) {
			strcat(command,"*");
            sprintf(params,"X'%04x'",currAddr+relOfs);
          } else
            sprintf(params,"[*%04xH]",currAddr+relOfs/*-2*/);
          sprintf(addrInfo," (ofs: %d=0x%02x)",relOfs,relOfs);
          break;
        case o_jm3:           // m=3: indirect relative
          relOfs = geti8();
          if (oldFmt) {
            strcat(command,"*");
            sprintf(params,"X'%04x'",currAddr+relOfs);
          } else
            sprintf(params,"[%04xH]",currAddr+relOfs/*-2*/);
          sprintf(addrInfo," (ofs: %d=0x%02x)",relOfs,relOfs);  
          break;
        case o_m4:           // m=4: indexed
          if (oldFmt) 
            strcat(command,"-");
          else
            strcpy(params,"[X]");
          break;
        case o_jm4:           // m=4: indexed
          if (oldFmt) 
            strcat(command,"-");
          else
            strcpy(params,"X");
          break;
        case o_m5:           // m=5: indexed with bias
          param = get8();
          if (oldFmt) {
			strcat(command,"+");
            sprintf(params,"X'%02x'",param);
          } else
            sprintf(params,"[X+%02xH]",param);
          break;
        case o_jm5:           // m=5: indexed with bias
          param = get8();
          if (oldFmt) {
			strcat(command,"+");
			sprintf(params,"X'%02x'",param);
		  } else
            sprintf(params,"X+%02xH",param);
          break;
        case o_m6:           // m=6: extended address
          param = get16();
          if (oldFmt) {
            strcat(command,"/");
            sprintf(params,"X'%04x'",param);
          } else {
            sprintf(params,"[%04xH]",param);
          } 
          break;
        case o_jm6:           // m=6: extended address
          param = get16();
          if (oldFmt) {
            strcat(command,"/");
            sprintf(params,"X'%04x'",param);
          } else {
            if (param & 0x8000)
              sprintf(params,"X+%04xH",param & 0x7fff);
            else
              sprintf(params,"%04xH",param);
          }
          break;
        case o_m7:           // m=7: indirect extended address
        case o_jm7:
          if (oldFmt)
            strcat(command,"=");
          param = cmd & 0xf8;  // mask out m0..7
          if ((param == OP_JMP) | (param == OP_RETJMP)) {
              param = get16();
              if (oldFmt) {
				strcat(command,"/");
				sprintf(params,"X'%04x'",param);
			  } else
              if (param & 0x8000) {  // bit7 = X+
                  sprintf(params,"[X+%04xH]",param & 0x7fff);
              } else {
                  sprintf(params,"[%04xH]",param);
              }
          } else {
              wordLength = 2;  // default to 2 byte for opcodes without "varlen" flag
              if (MD1600_opcdTab[(short)currByte].varlen)
                  // this may be 1,2,3 or 4 byte and depends on the word lenth set
                  wordLength = currWordLength;
              switch (wordLength) {
                case 1:
                  param = get8();
                  if (oldFmt) sprintf(params,"X'%02x'",param);
                  else sprintf(params,"%02xH",param);
                  break;
                case 2:
                  param = get16();
                  if (oldFmt) sprintf(params,"X'%04x'",param);
                  else sprintf(params,"%04xH",param);
                  break;
                case 3:
                  param = get16(); param2 = get8();
                  if (oldFmt) sprintf(params,"X'%04x%02x'",param,param2);
                  else sprintf(params,"%04x%02xH",param,param2);
                  break;
                case 4:
                  param = get16(); param2 = get16();
                  if (oldFmt) sprintf(params,"X'%08x'",param);
                  else sprintf(params,"%04x%04xH",param,param2);
                  break;
              }
          }
          break;
        case o_IBM_OBM:
          param = get8();
          param2 = get16();
              if (param & 0x8000) {  // bit7 = X+
                  sprintf(params,"%0x02,[X+%04xH]",param,param2 && 0x7fff);
              } else {
                  sprintf(params,"%0x02,[%04xH]",param,param2);
              }
          break;
        case o_inout:
          param = get8();
          param2 = (param & 0xe0) >> 5;
          param1 = param & 0x1f;
          if (oldFmt) {
			  sprintf(params,"X'%02x',X'%02x'",param2,param1);
		  } else {
			  sprintf(params,"%02xH,%02xH",param2,param1);
          }
          sprintf(addrInfo," (0x%02x)",param);
          break;
          
    }
    strupper(prefix); strupper(params);
    if (assemblerOut)
		printf("         %-6s %-20s %s%s\n",command,params,comment,addrInfo);
	else
		printf("%-17s %-6s %-20s %s%s\n",prefix,command,params,comment,addrInfo);
    linesOut++;    
}


char skipNoise(char c) {
	while (c < '0') {
		if (atEof) return 0;
		c = nextChar();
	}
	return c;
}

// skip empty and/or comment lines in bf/microdata boot format
// comments not allowed on original format. I assume that the loader
// ignored empty lines
char skipEmptyLines(void) {
	char c;
	c = nextChar();
	while ((c < '0') || (c == ';')){
	  c = skipNoise(c);
	  if (atEof) return 0;
	  if (c == ';') { // comment line
		  c = nextChar();
		  while ((c != 0x0a) && (c != 0x1f)) {
			  c = nextChar();
			  if (atEof) return 0;
		 }
		 c = nextChar(); 
	  }
	}
	return c;
}


void processFile(char *fileName) {
    char orgAddr[9];
    int firstAddr = 1;
    
    inFile = fopen(fileName,"rb");
    if (!inFile) {
        printf("unable to open \"%s\"\n",fileName);
        errors++;
        return;
    }
    atEof = 0;
    nextIsOrg = 0;
    if (inputFmt == fmtBF) nextIsOrg = 1;
    currAddr = org;
    
    while (!atEof) {
        if (nextIsOrg) {
            nextIsOrg = 0;
            // 4 bytes for start address
            strcpy(orgAddr,"0x0000");
            orgAddr[2] = skipEmptyLines();
            if (!atEof){
              orgAddr[3] = nextChar();
              orgAddr[4] = nextChar();
              orgAddr[5] = nextChar();
              nextIsOrg = 0; currByte = nextByte();
              if (nextIsOrg) atEof=1;
            }
            if (!atEof) {
              org = hex2int(orgAddr);
              currAddr = org;
              //printf("org %04x\n",org);
              if (firstAddr) {
				  if (assemblerOut) {
					  printf(" .CPU MD1600\n");
					  printf(" ORG 0x%04x\n",currAddr);
				  }
			  }
              firstAddr = 0;
            }
       } else
         currByte = nextByte();
       //if (!((currByte == 0x1F) & (inputFmt == fmtBF)))
       if (!nextIsOrg)
           processOpcode();
    }
    fclose(inFile);
}
    
int main (int argc, char **argv) {


  int c,index;
  errors = 0;  
  for (index=0;index<255;index++) {
      if (MD1600_opcdTab[index].opcodeval != index) {
          printf("error in opcodetab @ %d (%02xh) : %02x\n",index,index,MD1600_opcdTab[index].opcodeval);
          errors++;
      }
  }
  if(errors) _exit(1);
  errors = 0;
  opterr = 0;
  
  static struct option long_options[] =
  {
    {"help",      no_argument,       0, 'h'},
    {"old",       no_argument,       0, 'l'},
    {"binary",    no_argument,       0, 'b'},
    {"address",   required_argument, 0, 'o'},
    {"nocomments",no_argument,       0, 'n'},
    {"asmout",    no_argument,       0, 'a'},
    {0, 0, 0, 0}
  };
  
  int option_index = 0;

  
  while ((c = getopt_long (argc, argv, "1234nbo:hl?a",long_options, &option_index)) != -1) {
    switch (c)
      {
	  case 1:
	    break;
      case '1':
        currWordLength = 1;
        break;
      case '2':
        currWordLength = 2;
        break;
      case '3':
        currWordLength = 3;
        break;
      case '4':
        currWordLength = 4;
        break;
      case 'b':
        inputFmt = fmtBin;
        break;
      case 'o':
        org = hex2int(optarg);
        break;
      case 'n':
        comments = 0;
        break;
      case 'l':
        oldFmt = 1;
        break;
      case 'a':
        assemblerOut = 1;
        break;
      default:
        usage(argv[0]);
      }
  }

  if (optind >= argc) {
      printf("no input files specified\n");
      _exit(1);
  }
  printf("\n");
  
  if (assemblerOut) putchar(';');
  printf(MYSELF,VERMAJ,VERMIN);
  
  for (index = optind; index < argc; index++)
    processFile (argv[index]);
  if (linesOut > 0)
    if (comments)
      if (oldFmt)
        printf("\n; Postfixes: * indirect(m2,3), - indexed(m4), + indexed bias(m5)\n; / extended address(m6), = indirect extended address(m7)\n");
  return (errors = 0);
}

